#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Nov 29, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : metrics.py
project  : nadl
license  : GPL-3.0+

* Metric class and functions
* Some simple metrics that not in optax loss.  For complex metrics, use sklearn.
"""
# ruff: noqa: F722

import operator
from abc import abstractmethod
from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
from equinox import (
  AbstractVar,
  Module,
  combine,
  field,
  filter_pure_callback,
  filter_vmap,
  is_array,
  partition,
  tree_at,
  tree_equal,
  tree_pformat,
)

from jaxtyping import Array, Int, Num
from typing import Any, Literal, Self

import sklearn.metrics as m

from .utils import filter_concat

type N = Num[Array, "#a"]


def convert(x: float | Num | None) -> N:
  """Convert to float."""
  if x is None:
    return jnp.asarray(jnp.nan).reshape(-1)
  if isinstance(x, jax.Array):
    return x.reshape(-1)
  return jnp.asarray(x).reshape(-1)


def roc_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> N:
  """Compute ROC."""
  return jax.pure_callback(
    m.roc_auc_score, jax.ShapeDtypeStruct((), jnp.float32), labels, preds
  )


def average_precision_score(
  labels: Num[Array, " A"],
  preds: Num[Array, " A"],
  average: Literal["micro", "macro"] = "macro",
) -> N:
  """Compute PR."""
  return jax.pure_callback(
    partial(m.average_precision_score, average=average),
    jax.ShapeDtypeStruct((), jnp.float32),
    labels,
    preds,
  )


def pr_auc_score(labels: Num[Array, " A"], preds: Num[Array, " A"]) -> N:
  """Compute PR."""
  precision, recall, _ = filter_pure_callback(
    m.precision_recall_curve,
    labels,
    preds,
    result_shape_dtypes=(
      jax.ShapeDtypeStruct((preds.shape[0] + 1,), jnp.float32),
      jax.ShapeDtypeStruct((preds.shape[0] + 1,), jnp.float32),
      jax.ShapeDtypeStruct((preds.shape[0],), jnp.float32),
    ),
  )
  return jax.pure_callback(
    m.auc, jax.ShapeDtypeStruct((), jnp.float32), recall, precision
  )


class AbstractMetric[**P](Module):
  """Abstract Metric."""

  name: AbstractVar[str | None]

  @abstractmethod
  def __or__(self, value: Self) -> Self:
    """Or."""
    raise NotImplementedError

  @abstractmethod
  def __add__(self, value: Self) -> Self:
    """Add."""
    raise NotImplementedError

  def best(self, *args: P.args, **kwgs: P.kwargs) -> Self:
    """Best value."""
    return self[self.best_idx(*args, **kwgs)]

  def best_idx(self, *args: P.args, **kwgs: P.kwargs) -> Int[Array, "1"]:
    """Best value."""
    raise NotImplementedError

  def __repr__(self) -> str:
    """Representation."""
    return tree_pformat(self, short_arrays=True)

  @abstractmethod
  def __getitem__(self, idx: int | Int[Array, "#b"] | slice) -> Self:
    """Get item."""
    raise NotImplementedError

  def to_tuple(self) -> tuple[str | None, Any]:
    """To dict."""
    raise NotImplementedError


class Metric(AbstractMetric):
  """Base Metric."""

  value: N = field(default=jnp.nan, converter=convert)
  order: Literal["max", "min"] = "max"
  name: str | None = None

  def __or__(self, value: Self) -> Self:
    """Or."""
    s1, _ = partition(self, lambda x: is_array(x) and (not jnp.isnan(x).any()))
    return combine(s1, value)

  def __add__(self, value: Self) -> Self:
    """Add."""
    if self.name != value.name:
      raise ValueError(f"Name not match: this {self.name=} != {value.name=}")
    if self.order != value.order:
      raise ValueError(f"Order not match: this {self.order=} != {value.order=}")
    return filter_concat([self, value])

  def _max(self) -> Int[Array, "1"]:
    """Best value."""
    return jnp.nanargmax(self.value)

  def _min(self) -> Int[Array, "1"]:
    """Worst value."""
    return jnp.nanargmin(self.value)

  def best_idx(self) -> Int[Array, "1"]:
    """Best value."""
    return self._max() if self.order == "max" else self._min()

  def __getitem__(self, idx: int | Int[Array, "#b"] | slice) -> Self:
    """Get item."""
    if self.value is None:
      raise ValueError("No value.")
    return tree_at(lambda x: x.value, self, self.value[idx])

  def to_tuple(self) -> tuple[str | None, N | float]:
    """To dict."""
    if self.value.ndim <= 1:
      return self.name, self.value.item()
    return self.name, self.value


class Accuracy(Metric):
  """Accuracy."""

  @classmethod
  def create(
    cls,
    target: Num[Array, "*b a"],
    pred: Num[Array, "*b a"],
    name: str = "accuracy",
  ) -> Self:
    """Create from data."""
    if target.shape != pred.shape:
      raise ValueError(f"Shape not match: {target.shape=} != {pred.shape=}")
    match target.ndim:
      case 1:
        return cls(value=jnp.mean(target == pred), name=name)
      case 2:
        return cls(value=jnp.mean(target == pred, axis=1), name=name)
      case _:
        raise ValueError(f"Unsupported shape: {target.shape=}")

  @classmethod
  def create_empty(cls, name: str = "accuracy") -> Self:
    """Create empty."""
    return cls(value=jnp.nan, name=name)


class GroupMetric(AbstractMetric):
  """Group Metric."""

  metrics: list[AbstractMetric]
  name: str | None = None

  def __or__(self, value: Self) -> Self:
    """Or."""
    if tree_equal(*jax.tree.map(jnp.shape, [self.metrics, value.metrics])):
      s1, _ = partition(self, lambda x: is_array(x) and (not jnp.isnan(x).all()))
      return combine(s1, value)
    raise ValueError("Shape not match.")

  def __add__(self, value: Self) -> Self:
    """Add."""
    if self.name != value.name:
      raise ValueError(f"Name not match: this {self.name=} != {value.name=}")
    return filter_concat([self, value])

  def __getitem__(self, idx: int | Array | slice) -> Self:
    """Get item."""
    return tree_at(
      lambda x: x.metrics, self, jax.tree.map(operator.itemgetter(idx), self.metrics)
    )

  def __repr__(self) -> str:
    """Representation."""
    return tree_pformat(self.to_tuple(), short_arrays=True)

  def best_idx(
    self, which: int | Callable[[list[AbstractMetric]], Metric]
  ) -> Int[Array, "1"]:
    """Best value."""
    match which:
      case Callable():
        return which(self.metrics).best_idx()
      case int():
        return self.metrics[which].best_idx()
      case _:
        raise TypeError("which should be int or list[m] -> m.")

  def to_tuple(self) -> tuple[str | None, list]:
    """To dict."""
    return self.name, [m.to_tuple() for m in self.metrics]


class AccRocPR(GroupMetric):
  """Accuracy, ROC, PR."""

  @classmethod
  def create(
    cls,
    target: Num[Array, "*b a"],
    pred: Num[Array, "*b a"],
    name: str = "accrocpr",
  ) -> Self:
    """Create from data."""
    if target.shape != pred.shape:
      raise ValueError(f"Shape not match: {target.shape=} != {pred.shape=}")
    match target.ndim:
      case 1:
        return cls(
          metrics=[
            Metric(value=jnp.mean(target == pred), name="acc"),
            Metric(value=roc_auc_score(target, pred), name="roc"),
            Metric(value=pr_auc_score(target, pred), name="pr"),
            Metric(value=average_precision_score(target, pred), name="ap"),
          ],
          name=name,
        )
      case 2:
        return cls(
          metrics=[
            Metric(value=jnp.mean(target == pred, axis=1), name="acc"),
            Metric(value=jax.vmap(roc_auc_score)(target, pred), name="roc"),
            Metric(value=filter_vmap(pr_auc_score)(target, pred), name="pr"),
            Metric(value=filter_vmap(average_precision_score)(target, pred), name="ap"),
          ],
          name=name,
        )
      case _:
        raise ValueError(f"Unsupported shape: {target.shape=}")

  @classmethod
  def create_empty(cls, name: str = "accrocpr") -> Self:
    """Create empty."""
    return cls(
      metrics=[
        Metric(value=jnp.nan, name="acc"),
        Metric(value=jnp.nan, name="roc"),
        Metric(value=jnp.nan, name="pr"),
        Metric(value=jnp.nan, name="ap"),
      ],
      name=name,
    )


def dice_coef(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute dice coefficient."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred)

  return (2.0 * intersection) / (union + eps)


def iou_coef(y_true: jax.Array, y_pred: jax.Array, eps: float = 1e-8) -> jax.Array:
  """Compute intersection over union."""
  y_true = jnp.asarray(y_true)
  y_pred = jnp.asarray(y_pred)

  intersection = jnp.sum(y_true * y_pred)
  union = jnp.sum(y_true) + jnp.sum(y_pred) - intersection

  return intersection / (union + eps)


if __name__ == "__main__":
  print(
    GroupMetric(
      metrics=[AccRocPR.create_empty(name="rank"), AccRocPR.create_empty(name="comp")]
    )
  )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Dec  8, 2023
email    : Nasy <nasyxx+python@gmail.com>
filename : nets.py
project  : nadl
license  : GPL-3.0+

Pre define networks.
"""

from .blocks import DownSample, UpSample, ConvBlock, ResNetBasic, ResNetBottleneck

from flax import linen
import jax


class UNet(linen.Module):
  """UNet model for image segmentation."""

  out_channels: int
  num_features: int = 4
  init_features: int = 64
  kernel_size: int = 3
  use_depthwise: bool = False

  @linen.compact
  def __call__(self, x: jax.Array, train: bool = False) -> jax.Array:
    """Build the UNet model."""
    list_features = list(
      map(lambda n: self.init_features * (2**n), range(self.num_features))
    )
    skips = [
      x := DownSample(
        features=features,
        kernel_size=self.kernel_size,
        use_depthwise=self.use_depthwise,
        pool="before" if features > self.init_features else None,
      )(x, train=train)
      for features in list_features
    ]

    # Bottleneck
    x = ConvBlock(
      features=list_features[-1] * 2,
      kernel_size=self.kernel_size,
      use_depthwise=self.use_depthwise,
    )(
      linen.max_pool(x, window_shape=(2, 2), strides=(2, 2), padding="VALID"),
      train=train,
    )

    # Upsample path
    for features in reversed(list_features):
      skip = skips.pop()
      x = UpSample(
        features=features,
        kernel_size=self.kernel_size,
        use_depthwise=self.use_depthwise,
      )(x, skip, train=train)

    # Final layer
    return linen.Conv(features=self.out_channels, kernel_size=(1, 1))(x)


def resent18(out_channels: int, init_features: int = 64) -> ResNetBasic:
  """ResNet18."""
  return ResNetBasic(
    out_channels=out_channels, init_features=init_features, name="ResNet18"
  )


def resnet34(out_channels: int, init_features: int = 64) -> ResNetBasic:
  """ResNet34."""
  return ResNetBasic(
    out_channels=out_channels,
    init_features=init_features,
    num_blocks=(3, 4, 6, 3),
    name="ResNet34",
  )


def resnet50(out_channels: int, init_features: int = 64) -> ResNetBottleneck:
  """ResNet50."""
  return ResNetBottleneck(
    out_channels=out_channels, init_features=init_features, name="ResNet50"
  )


def resnet101(out_channels: int, init_features: int = 64) -> ResNetBottleneck:
  """ResNet101."""
  return ResNetBottleneck(
    out_channels=out_channels,
    init_features=init_features,
    num_blocks=(3, 4, 23, 3),
    name="ResNet101",
  )


def reset152(out_channels: int, init_features: int = 64) -> ResNetBottleneck:
  """ResNet152."""
  return ResNetBottleneck(
    out_channels=out_channels,
    init_features=init_features,
    num_blocks=(3, 8, 36, 3),
    name="ResNet152",
  )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

r"""Python ♡ Nasy.

    |             *         *
    |                  .                .
    |           .                              登
    |     *                      ,
    |                   .                      至
    |
    |                               *          恖
    |          |\___/|
    |          )    -(             .           聖 ·
    |         =\ -   /=
    |           )===(       *
    |          /   - \
    |          |-    |
    |         /   -   \     0.|.0
    |  NASY___\__( (__/_____(\=/)__+1s____________
    |  ______|____) )______|______|______|______|_
    |  ___|______( (____|______|______|______|____
    |  ______|____\_|______|______|______|______|_
    |  ___|______|______|______|______|______|____
    |  ______|______|______|______|______|______|_
    |  ___|______|______|______|______|______|____

author   : Nasy https://nasy.moe
date     : Apr 19, 2024
email    : Nasy <nasyxx+python@gmail.com>
filename : data.py
project  : nadl
license  : GPL-3.0+

Simple dataset and dataloader.
"""

from collections.abc import Callable, Iterator

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import field, tree_at
import numpy as np

from typing import NamedTuple

from .keys import Keys
from .loops import PG, RESC, PGThread


class DState[T](NamedTuple):
  """Dataloader state."""

  xs: T
  pad: jax.Array
  shape: tuple[int, ...]
  epoch: jax.Array = field(default_factory=lambda: jnp.array(0))
  step: jax.Array = field(default_factory=lambda: jnp.array(0))
  name: str | None = None


def _np_sort(x: jax.Array, axis: int | None = None) -> np.ndarray:
  """Sort the array."""
  return np.argsort(np.asarray(x), axis=axis)


@eqx.filter_jit
def fallback_argsort(x: jax.Array, axis: int | None = None) -> jax.Array:
  """Fallback to numpy argsort when CPU."""
  if jax.devices()[0].platform == "cpu":
    return jax.pure_callback(
      _np_sort, jax.ShapeDtypeStruct(x.shape, jnp.int32), x, axis
    )
  return x.argsort(axis=axis)


class IdxDataloader[T](eqx.Module):
  """Simple index dataloader."""

  length: int
  pad: int
  batch_size: int
  drop_num: int = 0
  transform: Callable[[jax.Array], T] = eqx.field(static=True, init=False)

  def __init__(
    self,
    length: int,
    batch_size: int,
    drop_last: bool = False,
    transform: Callable[[jax.Array], T] = lambda x: x,
  ) -> None:
    """Initiate the dataloader."""
    self.length = length

    if drop_last:
      self.drop_num = self.length % batch_size

    length = length if not drop_last else length - length % batch_size
    pad = (batch_size - r) % batch_size if (r := length % batch_size) else 0
    self.pad = pad = pad if pad != batch_size else 0

    self.batch_size = batch_size
    self.transform = eqx.filter_jit(transform)

  @eqx.filter_jit
  def __call__(self, key: jax.Array | None = None) -> DState[T]:
    """Get the indexes."""
    idxes = jnp.arange(self.length)
    if key is not None:
      idxes = jnp.take_along_axis(
        idxes,
        # NOTE: Fallback to numpy argsort since it has performance isssue in CPU.
        # https://github.com/google/jax/issues/10434
        fallback_argsort(jax.random.uniform(key, (self.length,))),
        axis=0,
      )
    length = self.length if not self.drop_num else self.length - self.drop_num

    idxes = jnp.r_[idxes, jnp.full(self.pad, -1, idxes.dtype)]
    idxes = idxes[: length + self.pad].reshape(-1, self.batch_size)
    return DState(
      self.transform(idxes), jnp.where(idxes == -1, 1, 0).astype(bool), idxes.shape
    )


def es_loop[T](
  loader: IdxDataloader[T],
  pg: PG,
  keys: Keys | None = None,
  epochs: int = 2,
  start_epoch: int = 1,
  prefix: str = "L",
  es: str = "E",
  ss: str = "S",
) -> Iterator[DState[T]]:
  """Simple epoch loop."""
  es, ss = f"{prefix}-{es}", f"{prefix}-{ss}"
  assert epochs > 0, "Epochs should be greater than 0."
  if keys:
    keys.reserve(epochs)

  vdl = eqx.filter_jit(eqx.filter_vmap(loader, axis_size=1))
  ds: DState[T] = vdl() if keys is None else vdl(keys(jnp.arange(epochs)))
  ds = tree_at(lambda d: d.name, ds, prefix, is_leaf=lambda x: x is None)

  if es in pg.tasks:
    pg.pg.reset(pg.tasks[es], total=epochs)
  else:
    pg.add_task(es, total=epochs, res="", visible=epochs > 1)
  pg.advance(pg.tasks[es], start_epoch - 1)
  if ss in pg.tasks:
    pg.pg.reset(pg.tasks[ss], total=ds.shape[0] * epochs, res="")
  else:
    pg.add_task(ss, total=ds.shape[0] * epochs, res="")
  pg.advance(pg.tasks[ss], (start_epoch - 1) * ds.shape[0])

  @eqx.filter_jit
  def _select(i: jax.Array, ii: jax.Array) -> DState[T]:
    return tree_at(
      lambda x: (x.epoch, x.step),
      jax.tree.map(lambda x: x[i, ii] if isinstance(x, jax.Array) else x, ds),
      (i + 1, i * ds.shape[0] + ii + 1),
    )

  with PGThread(pg.pg, pg.tasks[ss]) as pts, PGThread(pg.pg, pg.tasks[es]) as pte:
    for i in jnp.arange(start_epoch - 1, epochs):
      if epochs > 1:
        pte.completed += 1
      for ii in jnp.arange(ds.shape[0]):
        pts.completed += 1
        yield _select(i, ii)


def __test() -> None:
  """Test."""
  pg = PG.init_progress(extra_columns=(RESC,))
  keys = Keys.from_int_or_key(42)
  with pg:
    pg.console.print("Drop Last: False, Auto Pad: True")
    dl = IdxDataloader(314430, 256, drop_last=False)

    for i in es_loop(dl, pg, epochs=300, keys=keys, prefix="DFAT", start_epoch=10):
      pg.update_res(
        "DFAT-S", {"epoch": i.epoch.item(), "step": i.step.item(), "name": i.name}
      )
      continue
    pg.console.print(i)
    pg.console.print("Drop Last: True, Auto Pad: True")
    dl = IdxDataloader(10, 3, drop_last=True)
    for i in es_loop(dl, pg, keys, prefix="DTAT"):
      pg.console.print(i)


if __name__ == "__main__":
  __test()
